import unittest

from exodus_core.analysis.static_analysis import StaticAnalysis
from exodus_core.analysis.apk_signature import ApkSignature

import logging
logging.disable(logging.CRITICAL)

PHASH_SIZE = 16


def phash(apk):
    sa = StaticAnalysis(apk)
    return sa.get_icon_phash()


def save_icon(apk, path):
    sa = StaticAnalysis(apk)
    return sa.save_icon(path)


def icon_path(apk):
    sa = StaticAnalysis(apk)
    return sa.get_icon_path()


def list_classes(apk):
    sa = StaticAnalysis(apk)
    sa.load_trackers_signatures()
    return sa.get_embedded_classes()


def list_trackers(apk):
    sa = StaticAnalysis(apk)
    sa.load_trackers_signatures()
    return sa.detect_trackers()


def version_code(apk):
    sa = StaticAnalysis(apk)
    sa.load_trackers_signatures()
    return sa.get_version_code()


class TestExodus(unittest.TestCase):

    def test_trackers_list(self):
        sa = StaticAnalysis()
        sa.load_trackers_signatures()
        self.assertIsNotNone(sa.signatures)
        self.assertGreater(len(sa.signatures), 70)

    def test_icon_phash_greater_than_zero(self):
        apps = ['francetv']
        for app in apps:
            with self.subTest(app=app):
                icon_phash = phash('./apks/{}.apk'.format(app))
                self.assertGreater(icon_phash, 0)

    def test_icon_similarity_same_app(self):
        sa = StaticAnalysis()
        apps = ['braiar', 'whatsapp']
        for app in apps:
            with self.subTest(app=app):
                icon_phash = phash('./apks/{}.apk'.format(app))
                diff = sa.get_icon_similarity(icon_phash, icon_phash)
                self.assertEqual(diff, 1.0)

    def test_icon_similarity_different_app(self):
        sa = StaticAnalysis()
        phash_1 = phash('./apks/braiar.apk')
        phash_2 = phash('./apks/whatsapp.apk')
        diff_1 = sa.get_icon_similarity(phash_1, phash_2)
        diff_2 = sa.get_icon_similarity(phash_2, phash_1)

        self.assertEqual(diff_1, 0.5390625)
        self.assertEqual(diff_2, 0.5390625)

    def test_icon_similarity_different_similarities(self):
        sa = StaticAnalysis()
        phash_1 = phash('./apks/braiar.apk')
        phash_2 = phash('./apks/whatsapp.apk')
        phash_3 = phash('./apks/hsbc.apk')
        diff_1 = sa.get_icon_similarity(phash_1, phash_3)
        diff2 = sa.get_icon_similarity(phash_2, phash_3)
        self.assertNotEqual(diff_1, diff2)

    def test_app_uid(self):
        apps = [
            {'name': 'braiar', 'uid': '31BE732147F50EA10063BEACFAB2B4D6E0EEFC32'},
            {'name': 'whatsapp', 'uid': 'F799956E176E259FC28EB51AAD2E3519C9033619'},
            {'name': 'hsbc', 'uid': 'E3B4E87A002A37436CC6B008D3B43C0DB1A4FE13'},
            {'name': 'instapaper', 'uid': '64B99DBA34AFBC3709F75871837465892DF31771'},
            {'name': 'blooddonation', 'uid': '775AFB467E4EF556A42B28181A70B79BA67B4497'},
        ]
        for app in apps:
            with self.subTest(app=app['name']):
                uid = ApkSignature('./apks/{}.apk'.format(app['name'])).app_uid
                self.assertEqual(uid, app['uid'])

    def test_list_classes(self):
        apps = [
            {'name': 'braiar', 'nb_classes': 3896},
            {'name': 'whatsapp', 'nb_classes': 5457},
            {'name': 'hsbc', 'nb_classes': 7481},
            {'name': 'instapaper', 'nb_classes': 2676},
        ]
        for app in apps:
            with self.subTest(app=app['name']):
                classes = list_classes('./apks/{}.apk'.format(app['name']))
                self.assertIsNotNone(classes)
                self.assertEqual(len(classes), app['nb_classes'])

    def test_embedded_trackers(self):
        apps = [
            {'name': 'braiar', 'nb_trackers': 1},
            {'name': 'whatsapp', 'nb_trackers': 1},
            {'name': 'hsbc', 'nb_trackers': 3},
        ]
        for app in apps:
            with self.subTest(app=app['name']):
                apps = list_trackers('./apks/{}.apk'.format(app['name']))
                self.assertIsNotNone(apps)
                self.assertEqual(len(apps), app['nb_trackers'])

    def test_version_code(self):
        apps = [
            {'name': 'braiar', 'version': 13},
            {'name': 'whatsapp', 'version': 451987},
            {'name': 'hsbc', 'version': 61},
        ]
        for app in apps:
            with self.subTest(app=app['name']):
                version = version_code('./apks/{}.apk'.format(app['name']))
                self.assertIsNotNone(version)
                self.assertEqual(int(version), app['version'])


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestExodus)
    unittest.TextTestRunner(verbosity=2).run(suite)



import binascii
import itertools
import logging
import os
import re
import subprocess
import sys
import zipfile
from collections import namedtuple
from hashlib import sha256, sha1
from tempfile import NamedTemporaryFile, TemporaryDirectory

import dhash
import requests
import six
from androguard.core import axml
from androguard.core.apk import APK
from androguard.util import set_log
from cryptography.x509.name import ObjectIdentifier, _NAMEOID_DEFAULT_TYPE, _ASN1Type, NameAttribute
from PIL import Image

PHASH_SIZE = 8

set_log("INFO")


def which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def get_td_url():
    DEFAULT = "https://matlink.fr/token/email/gsfid"
    cred_paths_list = [
        'gplaycli.conf',
        os.path.expanduser("~") + '/.config/gplaycli/gplaycli.conf',
        '/etc/gplaycli/gplaycli.conf'
    ]
    config_file = None
    for filepath in cred_paths_list:
        if os.path.isfile(filepath):
            config_file = filepath
            break
    if config_file is None:
        return DEFAULT
    with open(config_file, mode='r') as c_file:
        for line in c_file:
            match = re.match(r'^token_url=(.*)$', line)
            if match:
                return match.group(1)
    return DEFAULT


def get_details_from_gplaycli(handle):
    """
    Get the details from gpc like creator, number of downloads, etc
    :param handle: application handle
    :return: application details dictionary
    """

    # FIXME: Returning None as Gplaycli with token is currently broken
    return None

    # TIME_BEFORE_RETRY = 2
    # API_SEARCH_LIMIT = 5
    #
    # gpc = gplaycli.GPlaycli()
    # gpc.token_enable = True
    # gpc.token_url = get_td_url()
    # try:
    #     gpc.token, gpc.gsfid = gpc.retrieve_token(force_new=False)
    # except (Exception, SystemExit):
    #     try:
    #         time.sleep(TIME_BEFORE_RETRY)
    #         gpc.token, gpc.gsfid = gpc.retrieve_token(force_new=True)
    #     except (Exception, SystemExit) as e:
    #         logging.error(e)
    #         return None
    # gpc.connect()
    # objs = gpc.api.search(handle, API_SEARCH_LIMIT)
    # try:
    #     for obj in objs:
    #         if obj['docId'] == handle:
    #             return obj
    #     return None
    # except Exception as e:
    #     logging.error('Unable to parse applications details')
    #     logging.error(e)
    #     return None


class Certificate:
    def __init__(self, cert):
        self.fingerprint = binascii.hexlify(cert.sha1).decode("ascii")
        self.issuer = cert.issuer.human_friendly
        self.subject = cert.subject.human_friendly
        self.serial = cert.serial_number

    def __str__(self):
        return 'Issuer: {} \n' \
               'Subject: {} \n' \
               'Fingerprint: {} \n' \
               'Serial: {}'.format(self.issuer, self.subject, self.fingerprint, self.serial)


class StaticAnalysis:
    def __init__(self, apk_path=None):
        self.apk = None
        self.apk_path = apk_path
        self.signatures = None
        self.compiled_tracker_signature = None
        self.classes = None
        self.app_details = None
        if apk_path is not None:
            self.load_apk()

    def _compile_signatures(self):
        """
        Compiles the regex associated to each signature, in order to speed up
        the trackers detection.
        :return: A compiled list of signatures.
        """
        self.compiled_tracker_signature = []
        try:
            self.compiled_tracker_signature = [re.compile(track.code_signature)
                                               for track in self.signatures]
        except TypeError:
            print("self.signatures is not iterable")

    def load_trackers_signatures(self):
        """
        Load trackers signatures from the official Exodus database.
        :return: a dictionary containing signatures.
        """
        self.signatures = []
        exodus_url = "https://reports.exodus-privacy.eu.org/api/trackers"
        r = requests.get(exodus_url)
        data = r.json()
        for e in data['trackers']:
            self.signatures.append(namedtuple('tracker', data['trackers'][e].keys())(*data['trackers'][e].values()))
        self._compile_signatures()
        logging.debug('{} trackers signatures loaded'.format(len(self.signatures)))

    def load_apk(self):
        """
        Load the APK file.
        """
        if self.apk is None:
            self.apk = APK(self.apk_path)

    @staticmethod
    def _get_embedded_classes(apkfile, depth=0):
        """
        Get the list of Java classes embedded into all DEX files.

        :return: set of Java classes names as string
        """
        if depth > 10:  # zipbomb protection
            logging.error(f'Max recursion depth in zip file reached: {apkfile}')
            return set()

        apk_regex = re.compile(r'.*\.apk')
        class_regex = re.compile(r'classes.*\.dex')
        classes = set()

        try:
            with TemporaryDirectory() as tmp_dir, zipfile.ZipFile(apkfile, 'r') as apk_zip:
                for info in apk_zip.infolist():
                    # apk files can contain apk files, again
                    if apk_regex.search(info.filename):
                        with apk_zip.open(info) as apk_fp:
                            classes = classes.union(StaticAnalysis._get_embedded_classes(apk_fp, depth + 1))

                    elif class_regex.search(info.filename):
                        if which('dexdump') is None:
                            logging.error("Unable to find dexdump executable, please install it.")
                            logging.error("On Debian-like OS, run sudo apt-get install dexdump")
                            sys.exit(1)

                        apk_zip.extract(info, tmp_dir)
                        run = subprocess.check_output(['dexdump', f'{tmp_dir}/{info.filename}'])
                        classes = classes.union(set(re.findall(r'[A-Z]+((?:\w+\/)+\w+)', run.decode(errors='ignore'))))
        except zipfile.BadZipFile as ex:
            logging.error(f'Unable to decode {apkfile}')
            raise Exception('Unable to decode the APK') from ex

        return classes

    def get_embedded_classes(self):
        """
        Get the list of Java classes embedded into all DEX files.
        :return: array of Java classes names as string
        """
        if self.classes is None:
            self.classes = list(self._get_embedded_classes(self.apk_path))

        return self.classes

    def detect_trackers_in_list(self, class_list):
        """
        Detect embedded trackers in the provided classes list.
        :return: list of embedded trackers
        """
        if self.signatures is None:
            self.load_trackers_signatures()

        def _detect_tracker(sig, tracker, class_list):
            for clazz in class_list:
                if sig.search(clazz):
                    return tracker
            return None

        results = []
        args = [(self.compiled_tracker_signature[index], tracker, class_list)
                for (index, tracker) in enumerate(self.signatures) if
                len(tracker.code_signature) > 3]

        for res in itertools.starmap(_detect_tracker, args):
            if res:
                results.append(res)

        trackers = [t for t in results if t is not None]
        logging.debug('{} trackers detected in {}'.format(len(trackers), self.apk_path))
        return trackers

    def detect_trackers(self, class_list_file=None):
        """
        Detect embedded trackers.
        :return: list of embedded trackers
        """
        if self.signatures is None:
            self.load_trackers_signatures()
        if class_list_file is None:
            return self.detect_trackers_in_list(self.get_embedded_classes())
        else:
            with open(class_list_file, 'r') as classes_file:
                classes = classes_file.readlines()
                return self.detect_trackers_in_list(classes)

    def get_version(self):
        """
        Get the application version name
        :return: version name
        """
        self.load_apk()
        return self.apk.get_androidversion_name()

    def get_version_code(self):
        """
        Get the application version code
        :return: version code
        """
        self.load_apk()
        return self.apk.get_androidversion_code()

    def get_permissions(self):
        """
        Get application permissions
        :return: application permissions list
        """
        self.load_apk()
        return self.apk.get_permissions()

    def get_app_name(self):
        """
        Get application name
        :return: application name
        """
        self.load_apk()
        return self.apk.get_app_name()

    def get_package(self):
        """
        Get application package
        :return: application package
        """
        self.load_apk()
        return self.apk.get_package()

    def get_libraries(self):
        """
        Get application libraries
        :return: application libraries list
        """
        self.load_apk()
        return self.apk.get_libraries()

    def get_icon_path(self):
        """
        Get the icon path in the ZIP archive
        :return: icon path in the ZIP archive
        """
        self.load_apk()
        return self.apk.get_app_icon()

    def get_application_details(self):
        """
        Get the application details like creator, number of downloads, etc.
        :param handle: application handle
        :return: application details dictionary
        """
        self.load_apk()

        if self.app_details is not None:
            return self.app_details

        details = get_details_from_gplaycli(self.get_package())
        if details is not None:
            self.app_details = details

        return details

    def _get_icon_from_details(self, path):
        """
        Get icon from applications details dictionary
        :param path: path where to write the icon file
        :return: icon path
        :raises Exception: if unable to find icon
        """
        details = self.get_application_details()
        if details is not None:
            for i in details.get('images'):
                if i.get('imageType') == 4:
                    f = requests.get(i.get('url'))
                    with open(path, mode='wb') as fp:
                        fp.write(f.content)
                    if os.path.isfile(path) and os.path.getsize(path) > 0:
                        return path

        raise Exception('Unable to download the icon from details')

    @staticmethod
    def _render_drawable_to_png(self, bxml, path):
        ap = axml.AXMLPrinter(bxml)
        print(ap.get_buff())

    def save_icon(self, path):
        """
        Extract the icon from the ZIP archive and save it at the given path
        :param path: destination path of the icon
        :return: destination path of the icon, None in case of error
        """
        try:
            icon = self.get_icon_path()
            if icon is None:
                raise Exception('Unable to get icon path')
            with zipfile.ZipFile(self.apk_path) as z:
                with open(path, 'wb') as f:
                    f.write(z.read(icon))
                with Image.open(path) as _:
                    logging.info('Get icon from APK: success')
                    return path
        except Exception:
            logging.warning('Unable to get the icon from the APK')
            return None

            # TODO: Set this back once details download is working again
            # logging.warning('Downloading icon from details')
            # try:
            #     saved_path = self._get_icon_from_details(path)
            #     logging.debug('Icon downloaded from application details')
            #     return saved_path
            # except Exception as e:
            #     logging.warning(e)

    def get_icon_phash(self):
        """
        Get the perceptual hash of the application icon
        :return: the perceptual hash, empty string in case of error
        """
        with NamedTemporaryFile() as ic:
            path = self.save_icon(ic.name)
            if path is None:
                logging.error('Unable to save the icon')
                return ''
            return self.get_phash(ic.name)

    @staticmethod
    def get_phash(image_name):
        """
        Get the perceptual hash of the given image
        :param image_name: name of the image file
        :return: the perceptual hash, empty string in case of error
        """
        dhash.force_pil()  # Force PIL

        try:
            image = Image.open(image_name).convert("RGBA")
            row, col = dhash.dhash_row_col(image, size=PHASH_SIZE)
            return row << (PHASH_SIZE * PHASH_SIZE) | col
        except IOError as e:
            logging.error(e)
            return ''

    @staticmethod
    def get_icon_similarity(phash_origin, phash_candidate):
        """
        Get icons similarity score [0,1.0]
        :param phash_origin: original icon
        :param phash_candidate: icon to be compared
        :return: similarity score [0,1.0]
        """
        diff = dhash.get_num_bits_different(phash_origin, phash_candidate)
        return 1 - 1. * diff / (PHASH_SIZE * PHASH_SIZE * 2)

    def get_application_universal_id(self):
        parts = [self.get_package()]
        for c in self.get_certificates():
            parts.append(c.fingerprint.upper())
        return sha1(' '.join(parts).encode('utf-8')).hexdigest().upper()

    def get_certificates(self):
        certificates = []

        def _my_name_init(self, oid, value, _type):
            if not isinstance(oid, ObjectIdentifier):
                raise TypeError("oid argument must be an ObjectIdentifier instance.")
            if not isinstance(value, six.text_type):
                raise TypeError("value argument must be a text type.")
            if len(value) == 0:
                raise ValueError("Value cannot be an empty string")
            if _type is None:
                _type = _NAMEOID_DEFAULT_TYPE.get(oid, _ASN1Type.UTF8String)
            if not isinstance(_type, _ASN1Type):
                raise TypeError("_type must be from the _ASN1Type enum")
            self._oid = oid
            self._value = value
            self._type = _type
        NameAttribute.__init__ = _my_name_init

        signs = self.apk.get_signature_names()
        for s in signs:
            c = self.apk.get_certificate(s)
            cert = Certificate(c)
            certificates.append(cert)

        return certificates

    def get_apk_size(self):
        """
        Get the APK file size in bytes
        :return: APK file size
        """
        return os.path.getsize(self.apk_path)

    def get_sha256(self):
        """
        Get the sha256sum of the APK file
        :return: hex sha256sum
        """
        BLOCKSIZE = 65536
        hasher = sha256()
        with open(self.apk_path, 'rb') as apk:
            buf = apk.read(BLOCKSIZE)
            while len(buf) > 0:
                hasher.update(buf)
                buf = apk.read(BLOCKSIZE)
        return hasher.hexdigest()

    def save_embedded_classes_in_file(self, file_path):
        """
        Save list of embedded classes in file.
        :param file_path: file to write
        """
        with open(file_path, 'w+') as f:
            f.write('\n'.join(self.get_embedded_classes()))

    def print_apk_infos(self):
        """
        Print APK information
        """
        permissions = self.get_permissions()
        libraries = self.get_libraries()
        certificates = self.get_certificates()
        print("=== Information")
        print('- APK path: {}'.format(self.apk_path))
        print('- APK sum: {}'.format(self.get_sha256()))
        print('- App version: {}'.format(self.get_version()))
        print('- App version code: {}'.format(self.get_version_code()))
        print('- App UID: {}'.format(self.get_application_universal_id()))
        print('- App name: {}'.format(self.get_app_name()))
        print('- App package: {}'.format(self.get_package()))
        print('- App permissions: {}'.format(len(permissions)))
        for perm in permissions:
            print('    - {}'.format(perm))
        print('- App libraries:')
        for lib in libraries:
            print('    - {}'.format(lib))
        print('- Certificates: {}'.format(len(certificates)))
        for cert in certificates:
            print('    - {}'.format(cert))

    def print_embedded_trackers(self):
        """
        Print detected trackers
        """
        trackers = self.detect_trackers()
        print('=== Found trackers: {}'.format(len(trackers)))
        for t in trackers:
            print(' - {}'.format(t.name))


if __name__ == '__main__':
    apk_path = sys.argv[1]
    sa = StaticAnalysis(apk_path)
    sa.print_apk_infos()
    sa.print_embedded_trackers()



import requests
import os

EXODUS_LOGIN_URI = '/api/get_auth_token/'


class ExodusConnector:
    """
    εxodus connector helps you to interact with it.
    Usage:
    >>> exodus = ExodusConnector("http://localhost:8000", "/api/report/1/")
    >>> exodus.login("user", "password")
    >>> exodus.get_report_info()
    >>> exodus.download_apk("/tmp")
    >>> exodus.upload_pcap("/tmp/my_file.pcap")
    """
    def __init__(self, host, report_info_uri):
        """
        Instantiate a new εxodus connector.
        ``report_info_uri`` is displayed on top of each report
        when you are connected with a superuser account.
        :param host: εxodus host e.g. http://localhost:8000
        :param report_info_uri: report URI e.g. /api/report/1/
        """
        self.host = host
        self.report_info_uri = report_info_uri
        self.access_token = ''
        self.report_info = None

    def login(self, username, password):
        """
        Login with the provided credentials and retrieve an authentication token.
        :param username: username
        :param password: password
        """
        r = requests.post(
            '{}{}'.format(self.host, EXODUS_LOGIN_URI),
            json={'username': username, 'password': password}
        )
        ret_code = r.status_code
        if ret_code != 200:
            raise Exception('Unable to login')
        self.access_token = r.json()['token']

    def get_report_info(self):
        """
        Retrieve report information like report ID, APK download URL, ...

        {
          'report_id': 15,
          'creation_date': '2017-11-13T18:55:53.452028Z',
          'apk_dl_link': '/api/apk/15/',
          'handle':'fr.meteo',
        }

        :return: dictionary containing report information
        """
        if not self.access_token:
            raise Exception('User not authenticated')

        r = requests.get(
            '{}{}'.format(self.host, self.report_info_uri),
            headers={"Authorization": "Token {}".format(self.access_token)}
        )
        ret_code = r.status_code
        if ret_code != 200:
            raise Exception('Unable to get report info')

        self.report_info = r.json()
        return self.report_info

    def download_apk(self, destination):
        """
        Download the APK into the provided destination path.
        :param destination: path to store the downloaded APK
        :return: path of the APK e.g. /tmp/fr.meteo.apk
        """
        if not self.access_token:
            raise Exception('User not authenticated')

        if not self.report_info['apk_dl_link']:
            raise Exception('Unable to get APK download link')

        url = '{}{}'.format(self.host, self.report_info['apk_dl_link'])
        local_filename = '{}.apk'.format(self.report_info['handle'])
        local_path = os.path.join(destination, local_filename)
        r = requests.get(
            url,
            stream=True,
            headers={'Authorization': 'Token {}'.format(self.access_token)}
        )
        with open(local_path, 'wb') as f:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)
        ret_code = r.status_code
        if ret_code != 200:
            raise Exception('Unable to download the APK')
        return local_path

    def upload_pcap(self, pcap_file):
        """
        DEPRECATED
        Upload the given PCAP file. The file has be in the Wireshark PCAP format.
        Once uploaded, εxodus will automatically analyze it.
        :param pcap_file: path to the local PCAP file
        """
        url = '{}{}'.format(self.host, self.report_info['pcap_upload_link'])
        with open(pcap_file, 'rb') as f:
            r = requests.post(
                url,
                files={'file': f},
                headers={"Authorization": "Token {}".format(self.access_token), "Content-Disposition": "attachment; filename={}".format(os.path.basename(pcap_file))}
            )
            ret_code = r.status_code
            if ret_code != 200:
                raise Exception('Unable to upload the PCAP file')

    def upload_flow(self, flow_file):
        """
        DEPRECATED
        Upload the given PCAP file. The file has be in the MITMDump FLOW format.
        Once uploaded, εxodus will automatically analyze it (not implemented yet).
        :param flow_file: path to the local FLOW file
        """
        url = '{}{}'.format(self.host, self.report_info['flow_upload_link'])
        with open(flow_file, 'rb') as f:
            r = requests.post(
                url,
                files={'file': f},
                headers={"Authorization": "Token {}".format(self.access_token), "Content-Disposition": "attachment; filename={}".format(os.path.basename(flow_file))}
            )
            ret_code = r.status_code
            if ret_code != 200:
                raise Exception('Unable to upload the FLOW file')



# Partial code from https://github.com/cpu/AndroidObservatory

import hashlib
import os


def findAPKFiles(dirPath):
    return findFiles(dirPath, ('.APK', '.apk'))


def findCertFiles(dirPath):
    return findFiles(dirPath, ('.RSA', '.DSA'))


def findFiles(dirPath, extensions):
    matches = []
    for root, dirnames, filenames in os.walk(dirPath):
        for filename in filenames:
            if filename.endswith(extensions):
                matches.append(os.path.join(root, filename))
    return matches


def hash(f):
    return hashlib.sha1(f).hexdigest().upper()


def hexToBits(hexString):
    return len(bin(int(hexString, 16))[2:])


def stringToHex(hexString):
    return "".join([line.strip().replace(':', '') for line in hexString])



from exodus_core.analysis.static_analysis import StaticAnalysis


class ApkSignature:
    def __init__(self, apk_path):
        sa = StaticAnalysis(apk_path)
        self.analysis = sa
        self.apk_path = apk_path
        self.apk_size = sa.get_apk_size()
        self.apk_sha256 = sa.get_sha256()
        self.handle = sa.get_package()
        self.version_code = sa.get_version_code()
        self.version_name = sa.get_version()
        self.app_name = sa.get_app_name()
        self.icon_phash = sa.get_icon_phash()
        self.permissions = sa.get_permissions()
        self.app_uid = sa.get_application_universal_id()
        self.certificates = sa.get_certificates()

    def __str__(self):
        return 'APK: {0}\n' \
               'Size: {1}\n' \
               'SHA256: {2}\n' \
               'Handle: {3}\n' \
               'Version code: {4}\n' \
               'Version name: {5}\n' \
               'App name: {6}\n' \
               'Icon pHash: {7}\n' \
               'App UID: {8}'.format(
                    self.apk_path,
                    self.apk_size,
                    self.apk_sha256,
                    self.handle,
                    self.version_code,
                    self.version_name,
                    self.app_name,
                    self.icon_phash,
                    self.app_uid
                )

    def get_icons_similarity(self, candidate):
        return StaticAnalysis.get_icon_similarity(self.icon_phash, candidate.icon_phash)

    def get_name_similarity(self, candidate):
        import jellyfish

        return {
            # Phonetic distance
            'mra': jellyfish.match_rating_comparison(self.app_name.replace(' ', ''),
                                                     candidate.app_name.replace(' ', '')),
            # String distance
            'jaro': jellyfish.jaro_winkler(self.app_name, candidate.app_name),
        }

    def get_handle_similarity(self, candidate):
        import jellyfish
        return jellyfish.jaro_winkler(self.handle, candidate.handle)

    def get_version_name_similarity(self, candidate):
        import jellyfish
        return jellyfish.jaro_winkler(self.version_name, candidate.version_name)

    def compare(self, candidate):
        is_same_handle = self.handle == candidate.handle
        is_same_name = self.app_name == candidate.app_name
        is_same_version = self.version_code == candidate.version_code
        is_same_size = self.apk_size == candidate.apk_size
        is_same_hash = self.apk_sha256 == candidate.apk_sha256
        is_same_icon = int(self.get_icons_similarity(candidate)) == 1
        is_another_version = is_same_handle and is_same_name and not is_same_version
        is_same_apk = is_same_handle and is_same_name and is_same_version and is_same_size and is_same_hash
        is_perfect_scam = is_same_handle and is_same_name and is_same_version and is_same_size and not is_same_hash

        c = {
            'files': {
                'origin': self.apk_path,
                'candidate': candidate.apk_path,
            },
            'names': {
                'origin': self.app_name,
                'candidate': candidate.app_name,
                'similarity': self.get_name_similarity(candidate),
                'matching': is_same_name
            },
            'handles': {
                'origin': self.handle,
                'candidate': candidate.handle,
                'similarity': self.get_handle_similarity(candidate),
                'matching': is_same_handle
            },
            'icons': {
                'origin': self.icon_phash,
                'candidate': candidate.icon_phash,
                'similarity': self.get_icons_similarity(candidate),
                'matching': is_same_icon
            },
            'version_codes': {
                'origin': self.version_code,
                'candidate': candidate.version_code,
                'matching': is_same_version
            },
            'app_uid': {
                'origin': self.app_uid,
                'candidate': candidate.app_uid,
                'matching': self.app_uid == candidate.app_uid
            },
            'version_names': {
                'origin': self.version_name,
                'candidate': candidate.version_name,
                'matching': self.version_name == candidate.version_name,
                'similarity': self.get_version_name_similarity(candidate),
            },
            'apk_hashes': {
                'origin': self.apk_sha256,
                'candidate': candidate.apk_sha256,
                'matching': is_same_hash
            },
            'result': {
                'is_same_apk': is_same_apk,
                'is_perfect_scam': is_perfect_scam,
            }
        }

        perceived_similarity = (3 * c['names']['similarity']['jaro'] + 2 * c['icons']['similarity'] + 1.5 *
                                c['handles']['similarity']) / 6.5
        if c['names']['similarity']['mra'] is not None:
            perceived_similarity = (3 * c['names']['similarity']['jaro'] + c['names']['similarity']['mra'] + 2 *
                                    c['icons']['similarity'] + 2 * c['handles']['similarity']) / 8.

        c['result']['perceived_similarity'] = perceived_similarity
        c['result']['is_scam'] = perceived_similarity > 0.75 and not is_another_version

        return c



#!/usr/bin/env python

import sys

from setuptools import setup, find_packages


if sys.version_info.major == 3 and sys.version_info.minor < 3:
    print("Unfortunately, your python version is not supported!\n Please upgrade at least to python 3.3!")
    sys.exit(1)

if sys.platform == 'darwin' or sys.platform == 'win32':
    print("Unfortunately, we do not support your platform %s" % sys.platform)
    sys.exit(1)

install_requires = [
    'androguard==4.1.0',
    'cryptography==42.0.4',
    'dhash==1.4',
    'jellyfish==0.5.6',
    'Pillow==10.2.0',
    'requests>=2.26,<2.32',
    'six==1.15.0',
    'traitlets==4.3.2'
]

setup(
    name='exodus_core',
    version='1.3.11',
    description='Core functionality of εxodus',
    author='Exodus Privacy',
    author_email='contact@exodus-privacy.eu.org',
    url='https://github.com/Exodus-Privacy/exodus-core',
    packages=find_packages(exclude=["*.tests", "*.tests.*", "test*", "tests"]),
    install_requires=install_requires,
    include_package_data=True,
    zip_safe=False,
)



# Partial code from https://github.com/cpu/AndroidObservatory

import re
import tempfile
from subprocess import check_output, check_call

from exodus_core.analysis.utils import stringToHex, hexToBits


class Certificate():
    def __init__(self, cert_file=None):
        self.certFile = None
        self.not_before = None
        self.not_after = None
        self.subject = None
        self.issuer = None
        self.serial = None
        self.fingerprint = None
        self.pubkey = None

        if cert_file:
            self.certFile = cert_file
            self.fromFile(cert_file)

    def __str__(self):
        return "X509 Cert. w/ Fingerprint: {0} Issuer: {1}".format(self.fingerprint, self.issuer)

    def fromFile(self, certFile):
        certInfo = check_output(["openssl", "pkcs7", "-inform", "DER",
                                 "-in", certFile, "-noout",
                                 "-print_certs", "-text"]).decode('utf-8')

        serialValParseA = re.search("^\s+Serial Number: \d+ (?:\(Negative\))?\(0x([0-9a-f:]+)\)$", certInfo,
                                    re.MULTILINE)
        serialValParseB = re.search("^\s+Serial Number:\s+(?:\(Negative\))?([0-9a-f:]+)$", certInfo, re.MULTILINE)
        issuerValParse = re.search("^\s+Issuer: (.+)$", certInfo, re.MULTILINE)
        subjectValParse = re.search("^\s+Subject: (.+)$", certInfo, re.MULTILINE)

        notBeforeValParse = re.search("^\s+Not Before: (.+)$", certInfo, re.MULTILINE)
        notAfterValParse = re.search("^\s+Not After : (.+)$", certInfo, re.MULTILINE)

        # Build X509 serial number
        if serialValParseA:
            self.serial = serialValParseA.group(1)
        elif serialValParseB:
            self.serial = stringToHex(serialValParseB.group(1))
        else:
            # e("Unable to parse X509 serial number from openssl.")
            raise Exception("Bad X509 serial number")

        # Build generic X509 information
        if issuerValParse and subjectValParse and notBeforeValParse and notAfterValParse:
            self.issuer = issuerValParse.group(1)
            self.subject = subjectValParse.group(1)
            self.not_before = notBeforeValParse.group(1)
            self.not_after = notAfterValParse.group(1)
        else:
            # e("Unable to parse generic X509 information from openssl.")
            raise Exception("Bad X509 fields")

        # Build RSA specific pubkey information
        if certInfo.find('rsaEncryption') != -1:
            pubkey = RSAKeyInfo()

            pubkeySizeValParse = re.search("^\s+(?:RSA Public Key|Public-Key): \((\d+) bit\)$", certInfo, re.MULTILINE)
            pubkeyModulusValParse = re.search("^\s+(?:Modulus|Modulus \([\d]+ bit\)):\s+([\s0-9a-f:]+)+", certInfo,
                                              re.MULTILINE)
            pubkeyExponentValParse = re.search("^\s+Exponent: (\d+) \(.*\)$", certInfo, re.MULTILINE)

            if pubkeySizeValParse and pubkeyModulusValParse and pubkeyExponentValParse:
                pubkey.bits = pubkeySizeValParse.group(1)
                pubkey.modulus = stringToHex(pubkeyModulusValParse.group(1))
                pubkey.exponent = pubkeyExponentValParse.group(1)
            else:
                # e("Unable to parse RSA Key information from openssl.")
                raise Exception("Bad RSA Key information")

        # Build DSA specific pubkey information
        elif certInfo.find('dsaEncryption') != -1:
            pubkey = DSAKeyInfo()

            pubkeyPubValParse = re.search("^\s+pub:\s+([\s0-9a-f:]+)+", certInfo, re.MULTILINE)
            pubkeyPValParse = re.search("^\s+P:\s+([\s0-9a-f:]+)+", certInfo, re.MULTILINE)
            pubkeyQValParse = re.search("^\s+Q:\s+([\s0-9a-f:]+)+", certInfo, re.MULTILINE)
            pubkeyGValParse = re.search("^\s+G:\s+([\s0-9a-f:]+)+", certInfo, re.MULTILINE)

            if pubkeyPubValParse and pubkeyPValParse and pubkeyQValParse and pubkeyGValParse:
                pubkey.p = stringToHex(pubkeyPubValParse.group(1))
                pubkey.q = stringToHex(pubkeyQValParse.group(1))
                pubkey.g = stringToHex(pubkeyGValParse.group(1))
                pubkey.pub = stringToHex(pubkeyPubValParse.group(1))

                pubkey.keybits = hexToBits(pubkey.p)
                pubkey.groupbits = hexToBits(pubkey.q)
            else:
                # e("Unable to parse DSA Key information from openssl.")
                raise Exception("Bad DSA Key information")
        else:
            # e("Unknown/Unsupported public key algorithm.")
            raise Exception("Unknown/Unsupported public key algorithm")

        self.pubkey = pubkey

        tmpFile = tempfile.NamedTemporaryFile()
        with tmpFile:
            # First extract a PEM certificate from the PKCS7 structure
            # into a temporary file (deleted at end of this with statement).
            # You can't directly access the X509 Certificate fingerprint using
            # the pkcs7 command, requiring these feats of hoop jumping.
            check_call(["openssl", "pkcs7", "-inform", "DER",
                        "-in", certFile, "-print_certs", "-out", tmpFile.name])

            # Then use the X509 command to get the fingerprint of the PEM certificate.
            certFPOutput = check_output(["openssl", "x509", "-inform", "PEM",
                                         "-in", tmpFile.name, "-noout", "-fingerprint"]).decode('utf-8')

            certFPValParse = re.search("^SHA1 Fingerprint=([0-9A-F:]+)$", certFPOutput)
            if not certFPValParse:
                # e("Unable to parse X509 certificate fingerprint")
                raise Exception("Bad X509 Certificate Fingerprint")

            self.fingerprint = stringToHex(certFPValParse.group(1))


class RSAKeyInfo():
    def __init__(self):
        self.algo = 'RSA'
        self.bits = None
        self.modulus = None
        self.exponent = None

    def __str__(self):
        return "{0} Bit RSA Pubkey".format(self.bits)


class DSAKeyInfo():
    def __init__(self):
        self.algo = 'DSA'
        self.keybits = None
        self.groupbits = None
        self.pub = None
        self.p = None
        self.q = None
        self.g = None

    def __str__(self):
        return "{0},{1} Bit DSA Pubkey".format(self.keybits, self.groupbits)
